      subroutine  eigen_trd( n, a, nma0, d_out, e_out, m0 )
!$    use OMP_LIB
      use MPI
      use eigen_devel
      implicit NONE

      integer                :: n, nma0, m0
      integer                :: nm, m
      real(8)                :: a(*)
      real(8)                :: d_out(1:n), e_out(1:n)

      include 'CSTAB.h'

      integer                :: nm0, nx0
      integer                :: i_1,j_1, i, j, k, j_2, j_3
      integer                :: nx, ierr

      integer, parameter     :: nm_max_L1 = 16*4
      integer, parameter     :: nm_max_L2 = 16*4*2


      real(8)                :: d1, d2
      real(8)                :: c1, c2

      integer :: lda, ldz, nz


      nm = nma0
      m  = m0

      call MPI_Barrier(TRD_COMM_WORLD,ierr)
      call eigen_timer_reset( BCAST=2, REDUCE=3, REDIST=2, GATHER=2 )

      lda = nma0

      call eigen_trd_(a, d_out, e_out, n, lda, m)


      call MPI_Barrier(TRD_COMM_WORLD,ierr)
      comm_time_reduction =
     &     eigen_timer_print( 'EigenExa (Tri-diagonalization)' )


      return
      end subroutine  eigen_trd

      subroutine  eigen_trd_(a,d_out,e_out,n,nm,m)
!$    use OMP_LIB
      use MPI
      use eigen_devel
      implicit NONE

      integer                :: n, nm, nmx, nv, nvx, m
      real(8)                :: a(nm,*)
      real(8)                :: d_out(1:n)
      real(8)                :: e_out(1:n)

      real(8), pointer       :: u_t(:), v_t(:), d_t(:)
      real(8), pointer       :: ux_vx(:)
      real(8), pointer       :: uy_vy(:)
      real(8), pointer       :: uvw_nn(:)

      include 'CSTAB.h'

      real(8)                :: d1,d2
      real(8)                :: d3,d4
      integer                :: nm_orig, nx, ny, n1, n2, n3
      integer                :: i,j,k

      integer                :: ierr, kx

      integer                :: offset1, offset2, offset3
      integer                :: offset4, offset5, offset6
      integer                :: offset7, offset8

      integer, parameter     :: nm_max_L1 = 16*4
      integer, parameter     :: nm_max_L2 = 16*4*2



      nx = (n-1)/x_nnod+1 +2
      ny = (n-1)/y_nnod+1 +2
      nv = nm
      call CSTAB_get_optdim(MAX(nx,ny), 6, nm_max_L1, nm_max_L2, nv)

      nx = (n-1)/x_nnod+1
      kx = y_nnod / n_common
      ny = (nx-1)/kx+1
      allocate(
     &     v_t(1:MAX(m*ny,    1, nv+6    )+n_columns),
     &     u_t(1:MAX(m*ny*kx, 6, nv+7+4*m)+n_columns),
     &     d_t(1:nv+n_columns),
     &     ux_vx(1:nv*2*m+2*n_columns),
     &     uy_vy(1:nv*2*m+2*n_columns),
     &     uvw_nn(m*3),
     &     stat=ierr)
      if ( ierr /= 0 ) then
         if ( TRD_inod == 1 ) print*,"Memory allocation error."
         call MPI_Abort(TRD_COMM_WORLD, 1, ierr)
      end if

      v_t = ZERO
      u_t = ZERO
      d_t = ZERO
      ux_vx = ZERO
      uy_vy = ZERO

      kx = nv*m+n_columns
      call CSTAB_adjust_base(ux_vx(1), a(1,1),offset1)
      call CSTAB_adjust_base(ux_vx(1+kx), a(1,1),offset3)
      call CSTAB_adjust_base(uy_vy(1), a(1,1),offset2)
      call CSTAB_adjust_base(uy_vy(1+kx), a(1,1),offset4)
      call CSTAB_adjust_base(u_t(1),a(1,1),offset5)
      call CSTAB_adjust_base(v_t(1),a(1,1),offset6)
      kx =                      !(L1_WINDOW/8)
!     &           +(L1_WINDOW)
     &     +(L1_LSIZE/8)
!     &           +(L1_LSIZE)
     &     +(L2_LSIZE/8)
      offset1 = offset1+kx*1
      offset2 = offset2+kx*3
      offset3 = offset3+kx*5
      offset4 = offset4+kx*2
      offset5 = offset5+kx*4
      offset6 = offset6+kx*6
      offset7 = offset7+kx*0
      call CSTAB_round_offset(offset1)
      call CSTAB_round_offset(offset2)
      call CSTAB_round_offset(offset3)
      call CSTAB_round_offset(offset4)
      call CSTAB_round_offset(offset5)
      call CSTAB_round_offset(offset6)
      call CSTAB_round_offset(offset7)

!!!   
      kx = nv*m+n_columns
      call MPI_Barrier(TRD_COMM_WORLD, ierr)

!$OMP PARALLEL
      call eigen_trd_body(a, nm, d_out, e_out, n, nv, m,
     &     ux_vx(1   +offset1), ! u_x(1:nv, 1:m)
     &     uy_vy(1   +offset2), ! u_y(1:nv, 1:m)
     &     ux_vx(1+kx+offset3), ! v_x(1:nv, 1:m)
     &     uy_vy(1+kx+offset4), ! v_y(1:nv, 1:m)
     &     u_t(1+offset5),      ! u_t(1:2*nv)
     &     v_t(1+offset6),      ! v_t(1:2*nv)
     &     d_t(1),              ! d_t(1:nv)
     &     uvw_nn(1),           ! u_nn(1:m)
     &     uvw_nn(1+1*m),       ! v_nn(1:m)
     &     uvw_nn(1+2*m))       ! w_nn(1:m)
!$OMP END PARALLEL

      call MPI_Barrier(TRD_COMM_WORLD, ierr)

      deallocate(v_t)
      deallocate(u_t)
      deallocate(d_t)
      deallocate(ux_vx)
      deallocate(uy_vy)
      deallocate(uvw_nn)


      return
      end subroutine            ! tred1_

      subroutine  eigen_trd_body(a, nm, d_out, e_out, n, nv, m_orig,
     &     u_x, u_y, v_x, v_y, u_t, v_t, d_t, u_nn, v_nn, w_nn)
!$    use omp_lib
      use MPI
#if 1
      use eigen_libs
#endif
      use eigen_devel
#if 1
      use eigen_house
#endif
      implicit NONE
      include 'CSTAB.h'

      integer, intent(IN)    :: n, nm, nv, m_orig
      real(8)                :: a(1:nm,*)
      real(8)                :: d_out(1:n), e_out(1:n)
      real(8)                :: u_t(1:2*nv), v_t(1:2*nv)
      real(8)                :: d_t(1:nv)
      real(8)                :: u_x(1:nv,*), u_y(1:nv,*)
      real(8)                :: v_x(1:nv,*), v_y(1:nv,*)
      real(8)                :: u_nn(*), v_nn(*), w_nn(*)

      integer, parameter     :: MBAND_ = 1

      real(8), save          :: c(MBAND_,MBAND_)
      real(8), save          :: u_n, v_n
      real(8)                :: t1, t2, t3
      integer                :: i, j, k, l

      integer                :: y_root, y_pos
      integer                :: x_root, x_pos
      integer                :: i_1, i_2, i_3, i_4
      integer                :: j_1, j_2, j_3, j_4
      integer                :: k_1, k_2, k_3, iref, iref_prev
      integer                :: l_1, l_2, l_3, l_4

      integer                :: m0, mm, m
      integer                :: i_block, i_base
      integer                :: local_rank, local_size
      logical                :: is_master
#if 1
      integer                :: loop_info(16)
#endif

      real(8)                :: d1,d2,dd(100)


#if TIMER_PRINT>1
      dd = ZERO
#endif
      m = MIN(m_orig, n)
!     
!     initialization
!     

      local_rank = 0
      local_size = 1
!$    local_rank = omp_get_thread_num()
!$    local_size = omp_get_num_threads()
      is_master = (local_rank == 0)

      loop_info(LOOP_INFO_L_RANK) = local_rank
      loop_info(LOOP_INFO_L_SIZE) = local_size

#if TIMER_PRINT>1
      if ( TRD_inod == 1 ) then
!$       if ( is_master ) then
            print*,"NUM.OF.PROCESS =",TRD_nnod,"(",x_nnod,y_nnod,")"
!$          print*,"NUM.OF.THREADS =",local_size
!$       endif
      endif
      call flush(6)
#endif

      call eigen_trd_init(
     &        a(1,1), nm, n,
     &        d_out(1), e_out(1),
     &        u_t(1), v_t(1), nv )

      mm = ((n-MOD(n,MBAND_))-1)/m+1

      do i_block=mm,MAX(1,3*(2-m)),-1

!$OMP BARRIER

         i_base = (i_block-1)*m
         m0     = MIN(m,n-i_base)

         call eigen_trd_load(
     &           a(1,1), nm, u_x(1,1), nv,
     &           d_t(1), v_x(1,1),
     &           m0, i_base, i_block )

!$       if ( is_master ) then

            k_2 = eigen_translate_g2l (i_base+m0, x_nnod, x_inod)
            k_3 = eigen_translate_g2l (i_base+m0, y_nnod, y_inod)
            flag_oncache = ( k_2 * k_3 / 2 * 2 < 5 * (L2_SIZE/8) )

!$       end if

!$OMP BARRIER

         k_2 = m0
         k_3 = MAX(1,3*(2-i_block))

         do k_1=k_2,k_3,-1

            iref      = k_1-MBAND_+1
            iref_prev = iref+1

            i = i_base+k_1


            loop_info(LOOP_INFO_X_START) =
     &           eigen_loop_start    (1,   x_nnod, x_inod)
            loop_info(LOOP_INFO_X_END  ) =
     &           eigen_loop_end      (i-1, x_nnod, x_inod)
            loop_info(LOOP_INFO_X_OINDX) =
     &           eigen_owner_index   (i-1, x_nnod, x_inod)
            loop_info(LOOP_INFO_X_ROOT ) =
     &           eigen_owner_node    (i-1, x_nnod, x_inod)
            loop_info(LOOP_INFO_X_POS  ) =
     &           eigen_translate_g2l (i-1, x_nnod, x_inod)

            loop_info(LOOP_INFO_Y_START) =
     &           eigen_loop_start    (1,   y_nnod, y_inod)
            loop_info(LOOP_INFO_Y_END  ) =
     &           eigen_loop_end      (i-1, y_nnod, y_inod)
            loop_info(LOOP_INFO_Y_OINDX) =
     &           eigen_owner_index   (i-1, y_nnod, y_inod)
            loop_info(LOOP_INFO_Y_ROOT ) =
     &           eigen_owner_node    (i-1, y_nnod, y_inod)
            loop_info(LOOP_INFO_Y_POS  ) =
     &           eigen_translate_g2l (i-1, y_nnod, y_inod)

            loop_info(LOOP_INFO_V_WIDTH ) = m0
            loop_info(LOOP_INFO_V_CURNT ) = k_1
            loop_info(LOOP_INFO_I_BASE  ) = i_base

!$          if ( is_master ) then
            if ( k_1 < k_2 ) then

#if TIMER_PRINT>1
               d1=eigen_get_wtime()
#endif
!     
!     w':= w-uv-vu
!     
               call eigen_trd_local_2update0(
     &                 u_x(1,iref),
     &                 u_x(1,iref_prev), v_x(1,iref_prev),
     &                 i-i_base, i+1, u_n, v_n )

#if TIMER_PRINT>1
               d2=eigen_get_wtime()
               dd(6)=dd(6)+(d2-d1)
#endif

               u_nn(1:k_2) = w_nn(1:k_2)

            end if

!     
!     u = Householder reflactor(...)
!     
#if TIMER_PRINT>1
            d1=eigen_get_wtime()
#endif

            call eigen_trd_u(
     &              d_out(i),
     &              u_x(1,iref), u_y(1,iref), nv,
     &              u_t(1), v_t(1), i,
     &              (k_2/=k_1), (k_1==k_3),
     &              v_x(1,iref_prev), v_y(1,iref_prev) )

#if TIMER_PRINT>1
            d2=eigen_get_wtime()
            dd(1)=dd(1)+(d2-d1)
#endif

!$          else
!$OMP BARRIER
!$          end if

#if TIMER_PRINT>1
!$          if ( is_master ) then
            d1=eigen_get_wtime()
!$          end if
#endif
!     
!     v:=Au
!     
            call eigen_trd_au(
     &              a(1,1), nm,
     &              u_x(1,iref), u_y(1,iref), v_x(1,iref),
     &              u_x(1, 1), v_x(1, 1), nv,
     &              u_t(1), v_t(1), d_t(1),
     &              loop_info,
     &              d_out(i), e_out(i), c(1,1),
     &              u_n, v_n, w_nn, v_nn )

!$          if ( local_size == 1 .OR.
!$   &           local_rank > 0 ) then
               if ( k_1 < k_2 ) then

                  call eigen_trd_local_2update(
     &                    u_x(1,1), nv,
     &                    u_x(1,iref_prev), v_x(1,iref_prev),
     &                    i-i_base-1, i+1, u_nn, v_nn,
     &                    local_size, local_rank )

               end if
!$          end if

!$OMP BARRIER

#if TIMER_PRINT>1
!$          if ( is_master ) then
            d2=eigen_get_wtime()
            dd(2)=dd(2)+(d2-d1)
!$          end if
#endif
!     
!     v=v-(UV+VU)u
!     v':= v-((u,v)/2|u|^2)u
!     
#if TIMER_PRINT>1
!$          if ( is_master ) then
            d1=eigen_get_wtime()
!$          end if
#endif

            call eigen_trd_vo(
     &              u_x(1,iref), v_x(1,iref), v_y(1,iref),
     &              u_x(1,1), v_x(1,1), nv,
     &              u_t(1), v_t(1), c(1,1), u_n, v_n,
     &              loop_info, (k_3/=k_1), v_nn )

#if TIMER_PRINT>1
!$          if ( is_master ) then
            d2=eigen_get_wtime()
            dd(4)=dd(4)+(d2-d1)
!$          end if
#endif

         end do                 ! k_1

!$OMP BARRIER

         if ( i_base == 0 ) then
            k_1=k_3; iref=k_1-MBAND_+1
            i = k_1

#if TIMER_PRINT>1
            d1=eigen_get_wtime()
#endif
!     
!     w':= w-uv-vu
!     
!$          if ( is_master ) then
            call reduce_dbl(v_nn, u_t, i-1, 1, x_COMM_WORLD)
!$          end if
!$OMP BARRIER

            call eigen_trd_local_2update(
     &              u_x(1,1), nv,
     &              u_x(1,iref), v_x(1,iref),
     &              i-i_base-1, i, w_nn, v_nn,
     &              local_size, local_rank )

!$OMP BARRIER
#if TIMER_PRINT>1
            if ( i_base == 0 ) then
               d2=eigen_get_wtime()
               dd(5)=dd(5)+(d2-d1)
            end if
#endif

         end if

         call eigen_trd_restore(
     &           a(1,1), nm, u_x(1,1), nv,
     &           d_t(1),
     &           m0, i_base )

!$OMP BARRIER

#if TIMER_PRINT>1
!$       if ( is_master ) then
         d1=eigen_get_wtime()
!$       end if
#endif
!     
!     A:=A-v^Tu-uT^v
!     

         if ( i_block > 1 ) then
            call eigen_2update(
     &              a(1,1), nm,
     &              u_x(1,1),u_y(1,1),v_x(1,1),v_y(1,1), nv,
     &              m0, i_base)
         end if

!$OMP BARRIER

#if TIMER_PRINT>1
!$       if ( is_master ) then
         d2=eigen_get_wtime()
         dd(3)=dd(3)+(d2-d1)
!$       end if
!$OMP BARRIER
#endif

      end do                    ! i_block

!$OMP BARRIER

!$    if ( local_rank < 2 ) then

      call eigen_trd_final(
     &        a(1,1), nm, n,
     &        d_out(1), e_out(1), u_t(1) )

!$    end if


#if TIMER_PRINT>1
      if(TRD_inod==1)then

!$OMP BARRIER
!$       if ( local_size == 1 .OR.
!$   &        local_rank == 1 ) then
            d_t(1) = dd(5)
!$       endif
!$OMP BARRIER
!$       if ( is_master ) then
         dd(5) = d_t(1)
!$       end if
!$OMP BARRIER

!$       if ( is_master ) then
         print*,"calc (u,beta)    ",dd(1)
         print*,"mat-vec (Au)     ",dd(2),dble(n)**3*2d-9/3/dd(2)
         print*,"2update (A-uv-vu)",dd(3),dble(n)**3*2d-9/3/dd(3)
         print*,"v=v-(UV+VU)u     ",dd(4)
         print*,"UV post reduction",dd(5)
!$       end if

      endif
      call flush(6)
#endif


      return
      end subroutine  eigen_trd_body

