      subroutine  eigen_trd_load(a, lda, u_x, ldv,
     &     d_t, buffer,
     &     m0, i_base, i_block)
!$    use omp_lib
      use eigen_devel
      use eigen_libs
      implicit NONE

      integer, intent(IN)    ::  lda, ldv, m0, i_base, i_block
      real(8), intent(INOUT) ::  a(1:lda, *)
      real(8), intent(OUT)   ::  u_x(1:ldv, *)
      real(8), intent(OUT)   ::  d_t(*)
      real(8), intent(OUT)   ::  buffer(*)

      integer                ::  x_pos, x_owner_nod
      integer                ::  y_pos, y_owner_nod
      integer                ::  i_1, i_2, i_3, i_4
      integer                ::  j_1, j_2, j_3, j_4
      integer                ::  k_1, k_2, k_3, k_4
      integer                ::  i, j, L, k0, i_root



      x_pos = eigen_translate_g2l(i_base+m0, x_nnod, x_inod)

      i_2 = eigen_loop_start(i_base+1,  y_nnod,y_inod)
      i_3 = eigen_loop_end  (i_base+m0, y_nnod,y_inod)
!$OMP DO
      do i_1=i_2,i_3
         j = eigen_translate_l2g(i_1, y_nnod,y_inod)
         u_x(1:x_pos, j-i_base) = a(1:x_pos, i_1)
      end do                 ! i_1
!$OMP END DO

      do j=1,m0,y_nnod
         k0 = MIN(m0+1-j,y_nnod)
         if ( k0 == y_nnod ) then
            i_root = eigen_owner_node (i_base+1, y_nnod, y_inod)
            i_1 = j + MOD(y_inod-i_root+y_nnod,y_nnod)
!$OMP MASTER
            call allgather_dbl( u_x(1, i_1), buffer, x_pos,
     &                          1, y_COMM_WORLD )
!$OMP END MASTER
!$OMP BARRIER
!$OMP DO
            do j_2=1,y_nnod
               j_3 = MOD(j_2-i_root+y_nnod,y_nnod)
               if ( j_3+1 /= y_inod ) then
                  k_2 = j_3*x_pos
                  u_x(1:x_pos, j+j_2-1) = buffer(k_2+1:k_2+x_pos)
               end if
            end do
!$OMP END DO
         else
!$OMP MASTER
            y_owner_nod = eigen_owner_node (i_base+j, y_nnod, y_inod)
            call bcastw_dbl( u_x(1, j), x_pos, y_owner_nod,
     &                       ldv, k0, buffer, 1, y_COMM_WORLD )
!$OMP END MASTER
         end if
      enddo

!$OMP BARRIER

      if ( diag_0 > 0 ) then
         i_2 = diag_0
         i_3 = eigen_loop_end  (i_base+m0-1, y_nnod,y_inod)
         if ( i_2 <= i_3 ) then
            i_4 = x_nnod/n_common
            j_2 = diag_1
            j_3 = eigen_loop_end  (i_base+m0-1, x_nnod, x_inod)
            j_4 = y_nnod/n_common
            k_2 = 0
            k_3 = (i_3-i_2)/i_4
!$OMP DO
            do  k_1 = k_2, k_3
               i_1 = i_2 + k_1 * i_4
               j_1 = j_2 + k_1 * j_4
               d_t(i_1)    = a(j_1, i_1)
               a(j_1, i_1) = ZERO
            end do           ! k_1
!$OMP END DO
         end if
      end if
#if 0
      if ( diag_0 > 0 ) then
         i_2 = diag_0
         i_3 = eigen_loop_end  (i_base+m0-1, y_nnod,y_inod)
         i_4 = x_nnod/n_common
         if ( y_nnod == x_nnod ) then
!DIR$ IVDEP
!DIR$ VECTOR ALWAYS
            do i_1=i_2,i_3,i_4
               d_t(i_1)   = a(i_1,i_1)
               a(i_1,i_1) = ZERO
            end do           ! i_1
         else
            do i_1=i_2,i_3,i_4
               j   = eigen_translate_l2g(i_1, y_nnod,y_inod)
               j_1 = eigen_translate_g2l(j,   x_nnod,x_inod)
               d_t(i_1)    = a(j_1, i_1)
               a(j_1, i_1) = ZERO
            end do           ! i_1
         end if
      end if
#endif


      return
      end subroutine  eigen_trd_load


      subroutine  eigen_trd_restore(a, lda, u_x, ldv,
     &     d_t,
     &     m0, i_base)
!$    use OMP_LIB
      use eigen_devel
      use eigen_libs
      implicit NONE

      integer, intent(IN)    ::  lda, ldv, m0, i_base
      real(8), intent(OUT)   ::  a(1:lda, *)
      real(8), intent(IN)    ::  u_x(1:ldv, *), d_t(*)

      integer                ::  i_1, i_2, i_3, i_4
      integer                ::  j_1, j_2, j_3, j_4
      integer                ::  k_1, k_2, k_3, k_4
      integer                ::  i, j, L


      i_2 = eigen_loop_start(i_base+1,  y_nnod,y_inod)
      i_3 = eigen_loop_end  (i_base+m0, y_nnod,y_inod)
!$OMP DO
      do i_1=i_2,i_3
         j   = eigen_translate_l2g(i_1, y_nnod,y_inod)
         j_3 = eigen_loop_end  (j, x_nnod,x_inod)
         a(1:j_3, i_1) = u_x(1:j_3, j-i_base)
      end do                    ! i_1
!$OMP END DO

      if ( diag_0 > 0 ) then
         i_2 = diag_0
         i_3 = eigen_loop_end  (i_base, y_nnod,y_inod)
         if ( i_2 <= i_3 ) then
            i_4 = x_nnod/n_common
            j_2 = diag_1
            j_3 = eigen_loop_end  (i_base, x_nnod, x_inod)
            j_4 = y_nnod/n_common
            k_2 = 0
            k_3 = (i_3-i_2)/i_4
!$OMP DO
            do  k_1 = k_2, k_3
               i_1 = i_2 + k_1 * i_4
               j_1 = j_2 + k_1 * j_4
               a(j_1, i_1) = d_t(i_1)
            end do              ! k_1
!$OMP END DO
         end if
      end if
#if 0
      if ( diag_0 > 0 ) then
         i_2 = diag_0
         i_3 = eigen_loop_end  (i_base, y_nnod,y_inod)
         i_4 = x_nnod/n_common
         if ( y_nnod == x_nnod ) then
            do i_1=i_2,i_3,i_4
               j_1 = i_1
               a(j_1, i_1) = d_t(i_1)
            end do              ! i_1
         else
            do i_1=i_2,i_3,i_4
!     j   = eigen_translate_l2g(i_1, y_nnod,y_inod)
!     j_1 = eigen_translate_g2l(j,   x_nnod,x_inod)
               j   = (i_1-1)*y_nnod+y_inod
               j_1 = (j-1)/x_nnod+1
               a(j_1, i_1) = d_t(i_1)
            end do              ! i_1
         end if
      end if
#endif

      return
      end subroutine  eigen_trd_restore

